# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import logging
import subprocess
from threading import Thread

import numpy as np
import zmq
from tornado import ioloop
from zmq.eventloop.zmqstream import ZMQStream
from zmq.utils.strtypes import b

allowed_zmq_patterns = {
    "REQ": zmq.REQ,
    "REP": zmq.REP,
    "PUB": zmq.PUB,
    "SUB": zmq.SUB,
}


class SocketCommunication:
    """ZMQ command_socket wrapper that offers multiple patterns and high-level interfaces to message serialisation."""

    context = None
    socket = None

    full_address = None
    pattern = "PUB"
    bind_bool = True
    subscribe_to = ""

    def __init__(
        self,
        protocol="tcp",
        address="127.0.0.1",
        port=12345,
        pattern="PUB",
        bind=True,
        subscribe_to="",
        auto_open=True,
    ):
        """
        :param protocol: protocols accepted by ZMQ, e.g. TCP
        :param address: IP address
        :param port: Port number
        :param pattern: Allowed values are: REQ, REP, PUB, SUB, while others might also work
        :param bind: If true, binds to address, else connects it
        :param subscribe_to: Subscription string for topic. If is list, subscribes to topics iteratively.
        """
        protocol = protocol.split(":")[
            0
        ]  # only first part is protocol and will be sufixed with ://

        if address.lower() == "localhost":
            address = "127.0.0.1"

        self.full_address = f"{protocol}://{address}:{port}"
        self.pattern = pattern.upper()
        self.bind_bool = bind
        self.subscribe_to = subscribe_to

        if auto_open:
            self.open()

    def open(self):
        logging.debug(
            f"{'Binding' if self.bind_bool else 'Connecting'} {self.pattern} socket on {self.full_address}."
        )

        self.context = zmq.Context()
        self.socket = self.context.socket(allowed_zmq_patterns.get(self.pattern))
        if self.bind_bool:
            self.socket.bind(self.full_address)
        else:
            self.socket.connect(self.full_address)

        if "SUB" in self.pattern:
            logging.debug(
                f"Subscribing to: '{self.subscribe_to}' on {self.full_address}"
            )
            if not isinstance(self.subscribe_to, list):
                self.socket.subscribe(b(self.subscribe_to))
            else:
                for topic in self.subscribe_to:
                    self.socket.subscribe(b(topic))

    def send_json(self, object=None):
        self.socket.send_json(obj=dict(object))

    def recv_json(self, *args, **kwargs):
        return self.socket.recv_json(*args, **kwargs)

    def recv_multipart(self):
        return self.socket.recv_multipart()

    def send_multipart(self, topic="", message=""):
        msg_parts = [b(part) for part in [topic, message]]
        return self.socket.send_multipart(msg_parts=msg_parts)

    def send_multipart_json(self, recipient="", message=None, flags=0):
        if not isinstance(message, dict):
            logging.warning(f"Message is type {type(message)}, but has to be dict.")
            return False

        self.socket.send(b(recipient), flags=flags | zmq.SNDMORE)
        return self.socket.send_json(obj=message, flags=flags)

    def send_array(self, array=None, metadata=None, flags=0, copy=True, track=False):
        """send a numpy array with metadata"""
        metadata_to_send = dict(
            dtype=str(array.dtype),
            shape=array.shape,
        )
        if isinstance(metadata, dict):
            metadata_to_send.update(metadata)

        self.socket.send_json(metadata_to_send, flags | zmq.SNDMORE)
        return self.socket.send(array, flags, copy=copy, track=track)

    def recv_array(self, flags=0, copy=True, track=False):
        """recv a numpy array"""
        received_metadata = self.socket.recv_json(flags=flags)
        data_segment = self.socket.recv(flags=flags, copy=copy, track=track)
        buf = memoryview(data_segment)
        received_array = np.frombuffer(buf, dtype=received_metadata["dtype"])
        return received_array.reshape(received_metadata["shape"])

    def recv_array_with_metadata(self, flags=0, copy=True, track=False):
        """recv a numpy array, plus additional metadata"""
        received_metadata = self.socket.recv_json(flags=flags)
        data_segment = self.socket.recv(flags=flags, copy=copy, track=track)
        buf = memoryview(data_segment)
        received_array = np.frombuffer(buf, dtype=received_metadata["dtype"])
        return received_metadata, received_array.reshape(received_metadata["shape"])

    def __del__(self):
        try:
            self.socket.close()
            self.context.term()
        except zmq.ZMQBaseError:
            logging.warning(
                f"ZMQBaseError on closing of command_socket/context of {self.full_address}"
            )


class ListenerStream(Thread):
    """Bundle ZMQ sockets and streams with shared event loop in thread for non-blocking callback."""

    socket_dict = {}
    stream_dict = {}
    recv_callback_input = None
    recv_callback_dict = {}
    ioloop = None
    daemon = True

    def __init__(self, socket_dict: dict, recv_callback_dict: (dict, None)):
        """
        :param socket_dict: dict of ZMQ command_socket objects
        :param recv_callback_dict: dict of functions to call on event on the sockets
        or a single function handle that will be used for each command_socket like the dict would be.
        """
        super(ListenerStream, self).__init__()

        self.socket_dict = socket_dict
        self.recv_callback_input = recv_callback_dict

        if self.recv_callback_input is None:
            logging.warning(
                "No stream recv_callbacks defined! Settings all callbacks to None"
            )
            for socket_name, _ in socket_dict.items():
                self.recv_callback_dict[socket_name] = None

        if callable(self.recv_callback_input):
            logging.debug(
                f"Only one callback defined. Setting to same: {recv_callback_dict}"
            )
            for socket_name, _ in socket_dict.items():
                self.recv_callback_dict[socket_name] = self.recv_callback_input

        for socket_name, socket_handle in socket_dict.items():
            new_stream = ZMQStream(socket=socket_handle)
            new_stream.on_recv(callback=self.recv_callback_dict[socket_name])
            self.stream_dict[socket_name] = new_stream

        self.ioloop = ioloop.IOLoop.instance()

    def run(self):
        self.ioloop.start()

    def stop(self):
        self.ioloop.stop()


def execute_in_commandline(cmd=None, return_std=False, **kwargs):
    if return_std:
        kwargs.update(
            {
                "stdout": subprocess.PIPE,
                "stderr": subprocess.PIPE,
            }
        )
    else:
        kwargs.update(
            {
                "stdout": subprocess.DEVNULL,
                "stderr": subprocess.DEVNULL,
            }
        )

    subprocess.Popen(cmd, **kwargs)



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import argparse
import logging
import time

import rpi_camera_colony
from rpi_camera_colony.config.config import load_config
from rpi_camera_colony.tools.comms import execute_in_commandline


class RemoteAcquisitionControl(object):
    """Manager to communicate with remote RPi. Instantiates PiAcquisitionControl on RPi."""

    instance_name = None
    control_socket_wrapper = None
    auto_init = None

    _connected = False
    _settings = None

    remote_python_interpreter = None
    remote_python_entrypoint = None

    def __init__(
        self,
        instance_name=None,
        config_data=None,
        control_socket_wrapper=None,
        auto_init=False,
        **kwargs,
    ):
        super(RemoteAcquisitionControl, self).__init__()

        if config_data is None and "config_file" not in kwargs:
            raise ValueError("Require either config_data object or config_file path.")

        self.instance_name = instance_name
        self.config_data = config_data or load_config(
            config_path=kwargs.get("config_file")
        )
        self.control_socket_wrapper = control_socket_wrapper

        self.remote_python_interpreter = self.config_data["general"].get(
            "remote_python_interpreter"
        )
        self.remote_python_entrypoint = self.config_data["general"].get(
            "remote_python_entrypoint"
        )
        self.remote_address = self.config_data["controllers"][self.instance_name].get(
            "address"
        )

        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)
                logging.debug(f"Setting attribute {attr} to {value}")

        if auto_init:
            self.initialise()

    @property
    def connected(self):
        return self._connected

    @connected.setter
    def connected(self, value):
        self._connected = value

        if self.connected:
            self.transmit_settings()
        else:
            self.pkill_remote()

    @property
    def config_data(self):
        return self._settings

    @config_data.setter
    def config_data(self, value=None):
        self._settings = value if hasattr(value, "keys") else None
        self.transmit_settings()

    def send_command(self, cmd_type="config", message_dict=None):
        message = {"type": cmd_type}
        message.update(message_dict)
        self.control_socket_wrapper.send_multipart_json(
            recipient=self.instance_name, message=message
        )

    def transmit_settings(self):
        if not self._settings:
            # logging.warning("Tried to transmit config_data, but cannot find any.")
            return

        if not self._connected:
            # logging.debug("Remote is closed.")
            return

        settings_for_this_instance = self.config_data["controllers"].get(
            self.instance_name
        )
        self.send_command(
            cmd_type="config",
            message_dict=settings_for_this_instance,
        )

    def _make_pi_command_base_list(self):
        username = self.config_data["general"].get("rpi_username", "pi")
        return [
            "ssh",
            username + "@" + self.remote_address,
            "sudo",
            "nohup",
        ]

    def initialise(self):
        if self.connected:
            logging.debug("Already initialised the remote acquisition.")
            return

        self.pkill_remote()
        time.sleep(1)
        cmd = self._make_pi_command_base_list() + [
            self.remote_python_interpreter,
            "-m",
            "rpi_camera_colony.acquisition",
            "--instance-name",
            self.instance_name,
            "--acquisition-name",
            self.config_data["general"]["acquisition_name"],
        ]
        execute_in_commandline(cmd=cmd)

        logging.debug(
            f"Initialised remote acquisition controller for: {self.remote_address} known as: {self.instance_name}."
        )

        self.connected = True

    def pkill_remote(self):
        cmd = self._make_pi_command_base_list() + ["pkill", "python"]
        execute_in_commandline(cmd=cmd)

    def start_acquisition(self):
        self.transmit_settings()
        time.sleep(1)
        self.send_command(cmd_type="command", message_dict={"status": "reset"})
        self.send_command(cmd_type="command", message_dict={"status": "start"})

    def stop_acquisition(self):
        self.send_command(cmd_type="command", message_dict={"status": "stop"})
        time.sleep(0.5)

    def cleanup(self):
        self.send_command(cmd_type="command", message_dict={"status": "close"})
        time.sleep(0.5)
        self.connected = False

    def __del__(self):
        self.cleanup()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()  # FIXME: check that not have to call connected=False before


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser_general = parser.add_argument_group("General")
    parser_general.add_argument(
        "--version", "-v", action="version", version=rpi_camera_colony.__version__
    )
    parser_ctrl = parser.add_argument_group("RemoteAcquisitionControl")
    parser_ctrl.add_argument(
        "--config_data-file",
        "-f",
        default="~/code/rpi_camera_colony/rpi_camera_colony/config/default.config",
        type=str,
        help="Path to config_data file.",
    )
    parser_ctrl.add_argument(
        "--instance-name",
        "-id",
        default="test_camera",
        type=str,
        help="Remote instance name to pick relevant section from config_data.",
    )
    parser_ctrl.add_argument(
        "--remote-python-interpreter",
        "-py",
        default="/home/pi/miniconda3/envs/py36/bin/python",
        type=str,
        help="Path to remote python interpreter. Can be simply 'python' for the default one.",
    )
    parser_ctrl.add_argument(
        "--command-socket",
        "-s",
        default=None,
        type=str,
        help="Socket address for command stream. "
        "Required on command line, but if object used in script, can also take zmq command_socket.",
    )
    parser_ctrl.add_argument(
        "--auto-init",
        "-i",
        default=False,
        type=bool,
        help="Automatically initialised the remote instance and communication channels.",
    )
    args = parser.parse_args()

    remote_acq_ctrl = RemoteAcquisitionControl(**vars(args))

    remote_acq_ctrl.initialise()
    remote_acq_ctrl.start_acquisition()
    time.sleep(10)
    remote_acq_ctrl.stop_acquisition()
    remote_acq_ctrl.cleanup()

    print("Exiting Remote Acq Ctrl.")



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import logging
import os
import time
from datetime import datetime


def get_datestr():
    return datetime.now().strftime("%Y%m%d_%H%M%S")


def make_recording_file_names(path=None, ext_sep=".", video_ext="h264"):
    basepath, _ = os.path.splitext(str(path))
    return {
        "video": ext_sep.join([basepath, "video", str(video_ext).replace(".", "")]),
        "ttl_out": ext_sep.join([basepath, "timestamps_ttl_out", "csv"]),
        "ttl_in": ext_sep.join([basepath, "timestamps_ttl_in", "csv"]),
        "metadata": ext_sep.join([basepath, "metadata", "json"]),
    }


def close_file_safe(file_handle):
    try:
        name = file_handle.name
    except BaseException:
        name = "not a file object"

    try:
        # Relevant actions
        file_handle.flush()
        os.fsync(file_handle.fileno())
        file_handle.close()
        file_handle = None
        # Report success
        logging.debug(f"Closed file:{name}")
    except BaseException:
        logging.debug(f"FAILED TO CLOSE FILE:{name}")



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import json
import logging
import os
import socket
import time
from pathlib import Path

from configobj import ConfigObj
from validate import Validator
from zmq.log.handlers import PUBHandler


def load_config(config_path=None, config_spec_path=None):
    if not config_path:
        raise ValueError("Need to specify a path for config_path.")

    config_path = Path(os.path.expanduser(config_path))
    if not config_path.exists():
        raise FileNotFoundError(f"File {config_path} not found.")

    if not config_spec_path:
        config_spec_path = Path(__file__).parent / "spec.config"

    config_spec_path = Path(os.path.expanduser(config_spec_path))
    if not config_spec_path.exists():
        raise FileNotFoundError(f"File {config_spec_path} not found.")

    config = ConfigObj(infile=str(config_path), configspec=str(config_spec_path))

    validation_success = config.validate(Validator(), copy=True)
    if not validation_success:
        logging.debug("Configuration file FAILED validation.")
    else:
        logging.debug(
            f"Config validation was successful for: \n {json.dumps(config.dict())}"
        )

    return config


def setup_logging(
    level="DEBUG", address_or_socket=None, root_topic="logging_root_topic"
):
    logging_handler = PUBHandler(interface_or_socket=address_or_socket)
    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s"
        " - %(processName)s %(filename)s:%(lineno)s"
        " - %(message)s"
    )
    formatter.datefmt = "%Y-%m-%d %H:%M:%S %p"
    logging_handler.setFormatter(formatter)
    logging_handler.root_topic = str(
        root_topic
    )  # ARM pyzmq version doesn't have root_topic as init argument

    logger = logging.getLogger()
    logger.setLevel(getattr(logging, level))
    logger.addHandler(logging_handler)
    logging.info("Logging initialised.")


def get_local_ip_address():
    """https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib"""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = "127.0.0.1"
    finally:
        s.close()
    return ip



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import json
import logging
from pathlib import Path

from rpi_camera_colony.config.config import get_local_ip_address
from rpi_camera_colony.tools.comms import ListenerStream
from rpi_camera_colony.tools.comms import SocketCommunication
from rpi_camera_colony.tools.files import get_datestr
from rpi_camera_colony.tools.files import make_recording_file_names

try:
    from rpi_camera_colony.acquisition.camera import Camera
except ImportError:
    logging.warning(
        "Will exit on error during PiAcquisitionControl init due to missing Camera import."
    )
    raise ImportError("Failed to import picamera")


class PiAcquisitionControl(object):
    active = True

    camera = None
    resolution = (640, 480)
    framerate = 90

    control_stream = None
    control_stream_ip = None
    control_stream_port = None

    instance_name = get_local_ip_address()
    data_path = "/home/pi/data/"
    acquisition_name = "test_recording"
    acquisition_file_base = None
    acquisition_files = None
    acquisition_settings = {}
    video_quality = 23

    def __init__(
        self,
        instance_name=None,
        data_path=None,
        acquisition_name=None,
        control_stream_ip=None,
        control_stream_port=None,
        **kwargs,
    ):
        super(PiAcquisitionControl, self).__init__()

        self.instance_name = instance_name or self.instance_name
        self.data_path = data_path or self.data_path
        self.acquisition_name = acquisition_name or self.acquisition_name
        self.control_stream_ip = control_stream_ip or self.control_stream_ip
        self.control_stream_port = control_stream_port or self.control_stream_port

        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)
                logging.debug(f"Set {self}, attr {attr} to value {value}")

        self.camera = Camera(**kwargs)
        self.camera.preview_static()

        # Make command_socket to receive control commands & subscribe to own channel
        self.control_socket = SocketCommunication(
            address=self.control_stream_ip,
            port=self.control_stream_port,
            pattern="SUB",
            bind=False,
            subscribe_to=self.instance_name,
        )
        self.control_stream = ListenerStream(
            socket_dict={1: self.control_socket.socket},
            recv_callback_dict=self._received_command,
        )
        self.control_stream.start()
        logging.debug("PiAcquisitionControl instantiated.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.camera is not None:
            if self.camera.preview is not None:
                self.camera.stop_preview()
            if self.camera.recording:
                self.camera.stop_recording()
            del self.camera
            self.camera = None

        self.control_stream.stop()
        logging.debug("Exiting PiAcquisitionControl.")

    def shutdown(self):
        """For use within contextmanager while loop: while this_object.active"""
        self.active = False

    def _write_metadata_file(self):
        # Note: Requires all variables to be public !
        metadata = {
            k: v
            for k, v in vars(self).items()
            if not isinstance(v, SocketCommunication)
            and not isinstance(v, ListenerStream)
            and not isinstance(v, type(self.camera))
        }

        save_path = self.acquisition_files["metadata"]
        with open(save_path, "w") as f:
            json.dump(metadata, f, indent=4, sort_keys=True)

        if not Path(save_path).exists():
            logging.error(f"Could not write metadata file to {save_path}")

    def _make_acquisition_paths(self):
        # Paths
        recording_name = "__".join(
            [self.acquisition_name, get_datestr(), self.instance_name]
        )
        acquisition_file_base = Path(self.data_path) / recording_name / recording_name
        acquisition_file_base.parent.mkdir(parents=True, exist_ok=True)
        if not acquisition_file_base.parent.exists():
            raise FileNotFoundError(
                f"Tried to make dir {str(acquisition_file_base.parent)}, but not found on check."
            )
        self.acquisition_file_base = str(acquisition_file_base)
        self.acquisition_files = make_recording_file_names(
            path=self.acquisition_file_base
        )
        logging.debug(f"{self.instance_name} - files: {self.acquisition_files}")

    def _received_command(self, command):
        command = [c.decode() for c in command]
        recipient, message_dict_str = command
        message = json.loads(message_dict_str, encoding="utf-8")

        logging.info(f"Recipient '{recipient}' reporting message '{message}'")

        if message["type"] == "command":
            self._update_camera_status(new_status=message["status"])
        elif message["type"] == "config":
            self._update_settings(config_data=message)

    def _update_settings(self, config_data=None):
        self.acquisition_settings = config_data

        for setting_name, setting_value in config_data.items():
            # Camera config_data
            if hasattr(self.camera, setting_name):
                setattr(self.camera, setting_name, setting_value)
                logging.debug(f"setattr(self.camera, {setting_name}, {setting_value})")

            # PiAcquisitionControl config_data
            if hasattr(self, setting_name):
                setattr(self, setting_name, setting_value)
                logging.debug(f"setattr(self, {setting_name}, {setting_value})")

    def _update_camera_status(self, new_status="stop"):
        logging.debug(f"New status: {new_status} on {self.instance_name}")

        if new_status in ["preview", "reset"]:
            self.camera.preview_static()

        elif new_status in "start":
            self._make_acquisition_paths()
            self._write_metadata_file()

            self.camera.start_recording(
                output_files=self.acquisition_files,
                format="h264",
                quality=self.video_quality,
            )

        elif new_status in "stop":
            self.camera.stop_recording()

        elif new_status in "close":
            self.shutdown()



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import logging
import sys
import time
from multiprocessing import Process
from multiprocessing import Queue

from rpi_camera_colony.control.conductor import Conductor
from rpi_camera_colony.control.conductor import parse_args_for_conductor


class ConductorAsProcess(Process):
    interrupt_queue = None
    conductor = None

    def __init__(self, controller_args=None, interrupt_queue=None):
        super(ConductorAsProcess, self).__init__()

        self.controller_args = controller_args
        self.interrupt_queue = interrupt_queue

        self.conductor = Conductor(**self.controller_args)

    def run(self) -> None:
        self.conductor.start_acquisition()

        while self.interrupt_queue.empty():
            time.sleep(1)

        self.conductor.stop_acquisition()
        self.conductor.cleanup()
        sys.exit(0)


if __name__ == "__main__":
    # Minimal example for sandboxing Conductor in separate process
    args_for_conductor = parse_args_for_conductor()
    interrupt_queue = Queue()  # Send value into this queue to stop the Conductor

    conductor_process = ConductorAsProcess(
        controller_args=args_for_conductor, interrupt_queue=interrupt_queue
    )
    conductor_process.run()

    while True:
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            logging.info("Keyboard interrupt. Cleaning up and exiting.")
            conductor_process.join()
            logging.info("Control processed joined and exiting. Good bye.")



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import logging
import time

try:
    import RPi.GPIO as GPIO
    import picamera
    from picamera import mmal
except ImportError:
    raise ImportError(
        "Can only run camera module on Raspberry Pi with RPi.GPIO and picamera packages installed."
    )
from rpi_camera_colony.tools.files import close_file_safe

GPIO.setwarnings(False)
GPIO.cleanup()
GPIO.setmode(GPIO.BOARD)


class VideoEncoder(picamera.PiVideoEncoder):
    """https://picamera.readthedocs.io/en/release-1.13/recipes2.html#custom-encoders"""

    frame_count = 0
    ttl_count = 0

    _ttl_out_pin = 8  # TODO: set to final value
    _ttl_out_duration = 0.001  # 1ms

    def __init__(self, *args, **kwargs):
        super(VideoEncoder, self).__init__(*args, **kwargs)

        for arg, val in kwargs.items():
            if hasattr(self, arg):
                setattr(self, arg, val)

    @property
    def ttl_out_pin(self):
        return self._ttl_out_pin

    @ttl_out_pin.setter
    def ttl_out_pin(self, value):
        self._ttl_out_pin = value
        if self.ttl_out_pin is not None:
            GPIO.setup(self.ttl_out_pin, GPIO.OUT, initial=GPIO.LOW)
            GPIO.output(self.ttl_out_pin, False)

    @property
    def ttl_out_duration(self):
        return self._ttl_out_duration

    @ttl_out_duration.setter
    def ttl_out_duration(self, value):
        self._ttl_out_duration = value

    def start(self, output, motion_output=None):
        super(VideoEncoder, self).start(output, motion_output)

    def close(self):
        super(VideoEncoder, self).close()

        if self.ttl_count != self.frame_count:
            logging.warning(
                f"Frame count ({self.frame_count}) and TTL count ({self.ttl_count}) do not match."
            )

    def _callback_write(self, buf, key=None):
        if (buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_FRAME_END) and not (
            buf.flags & mmal.MMAL_BUFFER_HEADER_FLAG_CONFIG
        ):
            if self.ttl_out_pin is not None:
                GPIO.output(self.ttl_out_pin, True)
                time.sleep(self.ttl_out_duration)
                GPIO.output(self.ttl_out_pin, False)

                self.ttl_count += 1

            self.parent._write_timestamps_frame_ttl_out(buf.pts, self.parent.timestamp)
            self.frame_count += 1

        return super(VideoEncoder, self)._callback_write(buf)


class Camera(picamera.PiCamera):
    _ttl_out_pin = 8
    _ttl_in_pin = 16
    _ttl_out_duration = 0.001

    file_timestamps_ttl_out = None
    file_timestamps_ttl_in = None

    clock_mode = "raw"

    def __init__(
        self,
        framerate=90,
        resolution=(640, 480),
        **kwargs,
    ):
        for attr, value in kwargs.items():
            if hasattr(self, attr):
                setattr(self, attr, value)
                logging.debug(f"Set {self}, attr {attr} to value {value}")

        super(Camera, self).__init__(
            framerate=framerate,
            resolution=resolution,
            clock_mode=self.clock_mode,
        )

    @property
    def ttl_out_pin(self):
        return self._ttl_out_pin

    @ttl_out_pin.setter
    def ttl_out_pin(self, value):
        self._ttl_out_pin = value

    @property
    def ttl_out_duration(self):
        return self._ttl_out_duration

    @ttl_out_duration.setter
    def ttl_out_duration(self, value):
        self._ttl_out_duration = value

    @property
    def ttl_in_pin(self):
        return self._ttl_in_pin

    @ttl_in_pin.setter
    def ttl_in_pin(self, value):
        self._ttl_in_pin = value

    def _get_video_encoder(self, camera_port, output_port, format, resize, **options):
        video_encoder = VideoEncoder(
            self, camera_port, output_port, format, resize, **options
        )
        video_encoder.ttl_out_pin = self.ttl_out_pin
        video_encoder.ttl_out_duration = self.ttl_out_duration
        return video_encoder

    def __del__(self):
        GPIO.cleanup()
        del self

    def _write_timestamps_ttl_in(self):
        if self.file_timestamps_ttl_in is not None:
            self.file_timestamps_ttl_in.write(f"{self.timestamp}\n")
            logging.debug(f"TTL-in detected at {self.timestamp}")

    def _write_timestamps_frame_ttl_out(self, cam_ts, frame_ts):
        if self.file_timestamps_ttl_out is not None:
            self.file_timestamps_ttl_out.write(f"{cam_ts},{frame_ts}\n")

    def start_recording(self, output_files=None, **kwargs):
        if self.ttl_out_pin is not None:
            # Open TTL file and write header
            self.file_timestamps_ttl_out = open(output_files["ttl_out"], "w")
            self.file_timestamps_ttl_out.write("# timestamp_frame, timestamp_ttl\n")

        if self.ttl_in_pin is not None:
            # Add event detection callback
            GPIO.setup(self.ttl_in_pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            GPIO.add_event_detect(
                self.ttl_in_pin, GPIO.RISING, self._write_timestamps_ttl_in
            )
            # Open TTL file and write header
            self.file_timestamps_ttl_in = open(output_files["ttl_in"], "w")
            self.file_timestamps_ttl_in.write("# timestamp_frame\n")

        super(Camera, self).start_recording(output=str(output_files["video"]), **kwargs)

    def stop_recording(self):
        if self.ttl_in_pin is not None:
            GPIO.remove_event_detect(self.ttl_in_pin)

        try:
            super(Camera, self).stop_recording()
        except BaseException:
            logging.error("CAMERA STOP EXCEPTION")

        # Close TTL files
        if self.file_timestamps_ttl_out is not None:
            close_file_safe(file_handle=self.file_timestamps_ttl_out)
        if self.file_timestamps_ttl_in is not None:
            close_file_safe(file_handle=self.file_timestamps_ttl_in)

    def preview_static(self, warmup_delay=3, alpha=255):
        if self.preview is not None:
            self.stop_preview()

        self.awb_mode = "auto"
        self.exposure_mode = "auto"

        self.start_preview(alpha=alpha)
        time.sleep(warmup_delay)

        awb_gains = self.awb_gains
        self.awb_mode = "off"
        self.awb_gains = awb_gains

        self.shutter_speed = self.exposure_speed
        self.exposure_mode = "off"

        logging.debug(
            f"Camera previewing & updated awb_gains={awb_gains}, shutter_speed={self.shutter_speed}"
        )



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import argparse
import logging
import time
from pathlib import Path

from rpi_camera_colony.acquisition.remote_control import RemoteAcquisitionControl
from rpi_camera_colony.config.config import load_config
from rpi_camera_colony.tools.comms import ListenerStream
from rpi_camera_colony.tools.comms import SocketCommunication
from rpi_camera_colony.tools.files import close_file_safe
from rpi_camera_colony.tools.files import get_datestr


def parse_args_for_conductor():
    parser = argparse.ArgumentParser(description="Input arguments")
    parser.add_argument(
        "--config-data-file",
        "-c",
        type=str,
        default=Path(__file__).parent.parent / "config/example.config",
        help="Settings file",
    )
    parser.add_argument(
        "--acquisition-name",
        "-n",
        type=str,
        default="default_acq_name_parser",
        help="Acquisition name [only name; path on Conductor should be defined in config]",
    )
    return parser.parse_args()


class Conductor(object):
    active = True

    config_data = None
    config_file = None
    acquisition_name = None

    _acquisition_controllers = {}
    acquiring = False

    auto_init = False
    auto_init_remote = True

    _logging_socket = None
    _logging_stream_callback = None
    _log_level = "INFO"
    _log_to_console = True
    _log_to_file = False
    _log_file = None

    _control_socket = None
    _comms_stream = None

    __cleaned_up = False

    def __init__(
        self,
        config_file=None,
        acquisition_name=None,
        logging_stream_callback=None,
        auto_init=False,
        auto_init_remote=True,
        delay_for_networking=4,
        delay_for_remote_instance=4,
    ):
        """ """
        super(Conductor, self).__init__()

        self.config_file = config_file
        self._load_config()

        self._log_level = self.config_data["log"]["level"]
        logger = logging.getLogger()
        logger.setLevel(getattr(logging, self._log_level))

        self.acquisition_name = acquisition_name
        self._logging_stream_callback = (
            logging_stream_callback or self._callback_receiver
        )
        self.auto_init = auto_init
        self.auto_init_remote = auto_init_remote

        if not self.acquisition_name:
            self.acquisition_name = self.config_data["general"].get(
                "acquisition_name", "default_acq_name_get"
            )
        else:
            self.config_data["general"]["acquisition_name"] = self.acquisition_name

        self._log_to_file = self.config_data["log"].get("log_to_file")
        self._log_to_console = self.config_data["log"].get("log_to_console")

        self._open_network_comms()
        logging.debug(f"Waiting {delay_for_networking}s for networking to come up..")
        time.sleep(delay_for_networking)

        self._make_acquisition_controllers(auto_init=auto_init_remote)
        logging.debug(
            f"Waiting {delay_for_remote_instance}s for remote instance to listen.."
        )
        time.sleep(delay_for_remote_instance)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def __del__(self):
        self.cleanup()

    def _load_config(self):
        self.config_data = load_config(config_path=self.config_file)

    def _open_network_comms(self):
        # Logging
        self._logging_socket = SocketCommunication(
            address=self.config_data["log"].get("address"),
            port=self.config_data["log"].get("port"),
            pattern="SUB",
            bind=True,
            auto_open=True,
        )
        # Control
        self._control_socket = SocketCommunication(
            address=self.config_data["control"].get("address"),
            port=self.config_data["control"].get("port"),
            pattern="PUB",
            bind=True,
            auto_open=True,
        )
        # Stream ioloop
        self._comms_stream = ListenerStream(
            socket_dict={
                "logging": self._logging_socket.socket,
                "control": self._control_socket.socket,
            },
            recv_callback_dict=self._callback_receiver,
        )
        self._comms_stream.start()

        if self._log_to_file is not None and self._log_to_file:
            log_file = ".".join(
                [self.config_data["log"].get("log_file"), get_datestr(), "log"]
            )
            self._log_file = open(log_file, "w")
            self._write_to_log(f"# Log for: {self.acquisition_name}\n")
            logging.debug(f"Logging  remote messages to {self._log_file}")

    def _callback_receiver(self, message=None):
        topic, message = [m.decode() for m in message]
        out_string = f"REMOTE: {topic} :: {message}".strip("\n")

        if self._log_to_console and self._log_level.upper() in topic.upper():
            logging.info(out_string)

        if (
            self._log_to_file is not None
            and self._log_to_file
            and self._log_file is not None
        ):
            self._write_to_log(out_string)

    def _write_to_log(self, out_string):
        if not self._log_file.closed:
            self._log_file.write(f"{out_string}\n")
            self._log_file.flush()

    def _make_acquisition_controllers(self, auto_init=True):
        """Make local objects to handle interaction with remote acquisition controller."""
        for instance_name, _ in self.config_data["controllers"].items():
            self._acquisition_controllers[instance_name] = RemoteAcquisitionControl(
                instance_name=instance_name,
                config_data=self.config_data,
                control_socket_wrapper=self._control_socket,
                auto_init=auto_init,
            )

    def initialise_acquisition_conductors(self):
        """Start up remote acquisition & transmit config_data in preparation for acquisition."""
        for _, acq in self._acquisition_controllers.items():
            acq.initialise()

    def start_acquisition(self):
        """Start acquisition."""
        for _, acq in self._acquisition_controllers.items():
            acq.start_acquisition()

        self.acquiring = True

    def stop_acquisition(self):
        """Stop acquisition."""

        if not self.acquiring:
            logging.info("Acquisition was not running.")

        for _, acq in self._acquisition_controllers.items():
            acq.stop_acquisition()

        self.acquiring = False

    def cleanup(self):
        """Post-acquisition tasks before exiting."""
        if self.__cleaned_up:
            return

        if (
            self._log_to_file is not None
            and self._log_to_file
            and self._log_file is not None
        ):
            close_file_safe(self._log_file)

        if self.acquiring:
            self.stop_acquisition()

        for _, acq in self._acquisition_controllers.items():
            acq.cleanup()

        self._acquisition_controllers = []
        if self._comms_stream is not None:
            self._comms_stream.stop()

        self.__cleaned_up = True
        self.active = False



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import sys
import time

from tqdm import tqdm

from rpi_camera_colony.control.conductor import Conductor
from rpi_camera_colony.control.conductor import parse_args_for_conductor


def main():
    args = parse_args_for_conductor()

    progress = tqdm(
        total=7200,
        desc="Recording progress",
        unit="seconds",
        bar_format="{desc}: {percentage:3.0f}%| | {n:.2f}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]",
    )

    conductor = Conductor(
        config_file=args.settings_file,
        acquisition_name=args.acquisition_name,
        auto_init=False,
    )
    # time.sleep(5)
    conductor.start_acquisition()

    while conductor.active:
        try:
            time.sleep(1)
            progress.update(n=1)
        except KeyboardInterrupt:
            print("\nKeyboard interrupt. Cleaning up and exiting.\n")
            conductor.stop_acquisition()
            conductor.cleanup()

    progress.close()
    sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import argparse
import logging
import platform
import sys
import time

import tqdm

import rpi_camera_colony
from rpi_camera_colony.config.config import get_local_ip_address
from rpi_camera_colony.config.config import setup_logging
from rpi_camera_colony.tools.comms import SocketCommunication


def parse_args_for_piacquisitioncontrol():
    parser = argparse.ArgumentParser()
    parser_general = parser.add_argument_group("General")
    parser_general.add_argument(
        "--print-interval",
        "-pi",
        default=0.500,
        type=float,
        help="Print interval [seconds, float]",
    )
    parser_general.add_argument(
        "--version", "-v", action="version", version=rpi_camera_colony.__version__
    )
    parser_acq_ctrl = parser.add_argument_group("PiAcquisitionControl")
    parser_acq_ctrl.add_argument(
        "--instance-name",
        "-id",
        default=get_local_ip_address(),
        type=str,
        help="Name of camera.",
    )
    parser_acq_ctrl.add_argument(
        "--data-path",
        "-p",
        default="/home/pi/data/",
        type=str,
        help="Path where to store all recordings.",
    )
    parser_acq_ctrl.add_argument(
        "--acquisition-name",
        "-n",
        default="test_recording",
        type=str,
        help="Base name for recording folder and files.",
    )
    parser_acq_ctrl.add_argument(
        "--max-acquisition-time",
        "-t",
        default=2 * 3600,
        type=int,
        help="Maximum recording time, seconds.",
    )
    parser_acq_ctrl.add_argument(
        "--log-ip",
        "-lip",
        default="192.168.100.10",
        type=str,
        help="IP address for logging.",
    )
    parser_acq_ctrl.add_argument(
        "--log-port", "-lport", default=55555, type=str, help="Port for logging."
    )
    parser_acq_ctrl.add_argument(
        "--log-level",
        "-ll",
        default="DEBUG",
        type=str,
        help="Logging level, e.g. INFO, DEBUG, etc.",
    )
    parser_acq_ctrl.add_argument(
        "--control-stream-ip",
        "-cip",
        default="192.168.100.10",
        type=str,
        help="IP address for control stream.",
    )
    parser_acq_ctrl.add_argument(
        "--control-stream-port",
        "-cport",
        default=54545,
        type=int,
        help="Port for control stream",
    )
    return parser.parse_args()


def main():
    args = parse_args_for_piacquisitioncontrol()

    # Exit if not on RPi -- makes parser outline available on non-RPi machines
    if "arm" not in platform.machine().lower():
        print("Not on Raspberry Pi. Exiting.")
        sys.exit(0)

    from rpi_camera_colony.acquisition.acquisition_control import PiAcquisitionControl

    # Set up logging & vars
    socket_wrapper = SocketCommunication(
        address=args.log_ip,
        port=args.log_port,
        pattern="PUB",
        bind=False,
        auto_open=True,
    )

    setup_logging(
        level=args.log_level,
        address_or_socket=socket_wrapper.socket,
        root_topic=args.instance_name,
    )

    max_time = args.max_acquisition_time
    print_interval = args.print_interval
    progress = tqdm.tqdm(
        total=max_time,
        desc="Recording progress",
        unit="seconds",
        bar_format="{desc}: {percentage:3.0f}%| | {n:.2f}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]",
    )

    # Run acquisition
    with PiAcquisitionControl(**vars(args)) as c:

        time.sleep(2)
        start_time = time.time()

        while (time.time() - start_time) < max_time and c.active:
            try:
                time.sleep(print_interval)
                print(f"Run time: {round(time.time() - start_time, 2)}")
                progress.update(n=print_interval)
            except KeyboardInterrupt:
                c.active = False
                break
            except BaseException:
                c.active = False

    progress.close()
    logging.info("Exiting main script.")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Author: Lars B. Rollik <L.B.Rollik@protonmail.com>
# License: BSD 3-Clause
import logging

level = "INFO"
logger = logging.getLogger()
logger.setLevel(getattr(logging, level))

formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s"
    " - %(processName)s %(filename)s:%(lineno)s"
    " - %(message)s"
)
formatter.datefmt = "%Y-%m-%d %H:%M:%S %p"

stream_handler = logging.StreamHandler()
stream_handler.setLevel(getattr(logging, level))
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

__version__ = "0.1.3.dev0"
__author__ = "Lars B. Rollik"
